@@ -2,10 +2,11 @@
 ######################################################################
 ### SNMP_util -- SNMP utilities using SNMP_Session.pm and BER.pm
 ######################################################################
-### Copyright (c) 1998-2005, Mike Mitchell.
+### Copyright (c) 1998-2007, Mike Mitchell.
 ###
 ### This program is free software; you can redistribute it under the
-### "Artistic License" included in this distribution (file "Artistic").
+### "Artistic License 2.0" included in this distribution
+### (file "Artistic").
 ######################################################################
 ### Created by:  Mike Mitchell   <Mike.Mitchell@sas.com>
 ###
@@ -43,7 +44,7 @@ use BER "1.02";
 use SNMP_Session "1.00";
 use Socket;
 
-$VERSION = '1.08';
+$VERSION = '1.12';
 
 @ISA = qw(Exporter);
 
@@ -364,6 +365,7 @@ sub encode_oid_with_errmsg ($);
 sub Check_OID ($);
 sub snmpLoad_OID_Cache ($);
 sub snmpQueue_MIB_File (@);
+sub MIB_fill_OID ($);
 
 sub version () { $VERSION; }
 
@@ -385,7 +387,7 @@ sub snmpopen ($$$) {
   # We can't split on the : character because a numeric IPv6
   # address contains a variable number of :'s
   my $opts;
- if( ($host =~ /^(\[.*\]):(.*)$/) || ($host =~ /^(\[.*\])$/) ) {
+ if( ($host =~ /^(\[.*\]):(.*)$/) or ($host =~ /^(\[.*\])$/) ) {
     # Numeric IPv6 address between []
     ($host, $opts) = ($1, $2);
   } else {
@@ -393,28 +395,28 @@ sub snmpopen ($$$) {
     ($host, $opts) = split(':', $host, 2);
   }
   ($port, $timeout, $retries, $backoff, $version, $v4onlystr) = split(':', $opts, 6)
-    if(defined($opts) && (length $opts > 0) );
+    if(defined($opts) and (length $opts > 0) );
 
-  undef($version) if (defined($version) && length($version) <= 0);
+  undef($version) if (defined($version) and length($version) <= 0);
   $v4onlystr = "" unless defined $v4onlystr;
   $version = '1' unless defined $version;
-  if (defined($port) && ($port =~ /^([^!]*)!(.*)$/)) {
+  if (defined($port) and ($port =~ /^([^!]*)!(.*)$/)) {
     ($port, $lhost) = ($1, $2);
     $nlhost = $lhost;
     ($lhost, $lport) = ($1, $2) if ($lhost =~ /^(.*)!(.*)$/);
-    undef($lhost) if (defined($lhost) && (length($lhost) <= 0));
-    undef($lport) if (defined($lport) && (length($lport) <= 0));
+    undef($lhost) if (defined($lhost) and (length($lhost) <= 0));
+    undef($lport) if (defined($lport) and (length($lport) <= 0));
   }
-  undef($port) if (defined($port) && length($port) <= 0);
-  $port = 162 if ($type == 1 && !defined($port));
+  undef($port) if (defined($port) and length($port) <= 0);
+  $port = 162 if ($type == 1 and !defined($port));
   $nhost = "$community\@$host";
   $nhost .= ":" . $port if (defined($port));
 
   if ((!defined($SNMP_util::Session))
-    || ($SNMP_util::Host ne $nhost)
-    || ($SNMP_util::Version ne $version)
-    || ($SNMP_util::LHost ne $nlhost)
-    || ($SNMP_util::IPv4only ne $v4onlystr)) {
+    or ($SNMP_util::Host ne $nhost)
+    or ($SNMP_util::Version ne $version)
+    or ($SNMP_util::LHost ne $nlhost)
+    or ($SNMP_util::IPv4only ne $v4onlystr)) {
     if (defined($SNMP_util::Session)) {
       $SNMP_util::Session->close();    
       undef $SNMP_util::Session;
@@ -459,11 +461,11 @@ sub snmpopen ($$$) {
       }
     }
     $SNMP_util::Session->set_timeout($timeout)
-      if (defined($timeout) && (length($timeout) > 0));
+      if (defined($timeout) and (length($timeout) > 0));
     $SNMP_util::Session->set_retries($retries)
-      if (defined($retries) && (length($retries) > 0));
+      if (defined($retries) and (length($retries) > 0));
     $SNMP_util::Session->set_backoff($backoff)
-      if (defined($backoff) && (length($backoff) > 0));
+      if (defined($backoff) and (length($backoff) > 0));
   }
   return $SNMP_util::Session;
 }
@@ -590,7 +592,7 @@ sub snmpwalk_flg ($$@) {
   #
   # Create/Refresh a reversed hash with oid -> name
   #
-  if (defined($hash_sub) && ($RevNeeded)) {
+  if (defined($hash_sub) and ($RevNeeded)) {
       %revOIDS = reverse %SNMP_util::OIDS;
       $RevNeeded = 0;
   }
@@ -624,7 +626,7 @@ sub snmpwalk_flg ($$@) {
   }
 
 
-  while(($SNMP_util::Version ne '1' && $session->{'use_getbulk'})
+  while(($SNMP_util::Version ne '1' and $session->{'use_getbulk'})
     ? $session->getbulk_request_response(0,
 					  $session->default_max_repetitions(),
 					  @nnoid)
@@ -644,7 +646,7 @@ sub snmpwalk_flg ($$@) {
       $ok = 0;
       my $tempo = pretty_print($oid);
       $noid = $avars[$ix];  # IlvJa
-      if ($tempo =~ /^$noid\./ || $tempo eq $noid ) {
+      if ($tempo =~ /^$noid\./ or $tempo eq $noid ) {
 	$ok = 1;
 	$upoid = $noid;
       } else {
@@ -661,7 +663,16 @@ sub snmpwalk_flg ($$@) {
       if ($ok) {
 	my $tmp = encode_oid_with_errmsg ($tempo);
 	return undef unless defined $tmp;
-	next if (exists($done{$tmp}));	# GIL
+	if (exists($done{$tmp})) {	# GIL, Ilvja
+	  #
+	  # We've detected a loop for $nnoid[$ix], so mark it as finished.
+	  # Exclude this variable from further requests.
+	  #
+	  $avars[$ix] = "";
+	  $nnoid[$ix] = "";
+	  $retvaltmprefs[$ix] = undef if $SNMP_util::Return_array_refs;
+	  next;
+	}
 	$nnoid[$ix] = $tmp;   # Keep on walking. (IlvJa)
 	my $tempv = pretty_print($value);
 	if (defined($hash_sub)) {
@@ -670,32 +681,32 @@ sub snmpwalk_flg ($$@) {
 	  #
 	  my $inst = "";
 	  my $upo = $upoid;
-	  while (!exists($revOIDS{$upo}) && length($upo)) {
+	  while (!exists($revOIDS{$upo}) and length($upo)) {
 	    $upo =~ s/(\.\d+?)$//;
-	    if (defined($1) && length($1)) {
+	    if (defined($1) and length($1)) {
 	      $inst = $1 . $inst;
 	    } else {
 	      $upo = "";
 	      last;
 	    }
 	  }	
-	  if (length($upo) && exists($revOIDS{$upo})) {
+	  if (length($upo) and exists($revOIDS{$upo})) {
 	    $upo = $revOIDS{$upo} . $inst;
 	  } else {
 	    $upo = $upoid;
 	  }
 
 	  $inst = "";
-	  while (!exists($revOIDS{$tempo}) && length($tempo)) {
+	  while (!exists($revOIDS{$tempo}) and length($tempo)) {
 	    $tempo =~ s/(\.\d+?)$//;
-	    if (defined($1) && length($1)) {
+	    if (defined($1) and length($1)) {
 	      $inst = $1 . $inst;
 	    } else {
 	      $tempo = "";
 	      last;
 	    }
 	  }	
-	  if (length($tempo) && exists($revOIDS{$tempo})) {
+	  if (length($tempo) and exists($revOIDS{$tempo})) {
 	    $var = $revOIDS{$tempo};
 	  } else {
 	    $var = pretty_print($oid);
@@ -987,7 +998,7 @@ sub toOID(@) {
   undef @retvar;
   foreach $var (@vars) {
     ($oid, $tmp) = &Check_OID($var);
-    if (!$oid && $SNMP_util::CacheLoaded == 0) {
+    if (!$oid and $SNMP_util::CacheLoaded == 0) {
       $tmp = $SNMP_Session::suppress_warnings;
       $SNMP_Session::suppress_warnings = 1000;
 
@@ -998,7 +1009,7 @@ sub toOID(@) {
 
       ($oid, $tmp) = &Check_OID($var);
     }
-    while (!$oid && $#SNMP_util::MIB_Files >= 0) {
+    while (!$oid and $#SNMP_util::MIB_Files >= 0) {
       $tmp = $SNMP_Session::suppress_warnings;
       $SNMP_Session::suppress_warnings = 1000;
 
@@ -1084,7 +1095,7 @@ sub snmpLoad_OID_Cache ($) {
     $txt = $1 if ($txt =~ /^[\'\"](.*)[\'\"]/);
     $oid = $1 if ($oid =~ /^[\'\"](.*)[\'\"]/);
     if (($txt =~ /^\.?\d+(\.\d+)*\.?$/)
-    &&  ($oid !~ /^\.?\d+(\.\d+)*\.?$/)) {
+    and  ($oid !~ /^\.?\d+(\.\d+)*\.?$/)) {
 	my($a) = $oid;
 	$oid = $txt;
 	$txt = $a;
@@ -1106,7 +1117,7 @@ sub Check_OID ($) {
   my($var) = @_;
   my($tmp, $tmpv, $oid);
 
-  if ($var =~ /^[a-zA-Z][\w\-]*(\.[a-zA-Z][\w\-])*/)
+  if ($var =~ /^[a-zA-Z][\w\-]*(\.[a-zA-Z][\w\-]*)*/)
   {
     $tmp = $&;
     $tmpv = $tmp;
@@ -1144,19 +1155,8 @@ sub snmpQueue_MIB_File (@) {
 #
 sub snmpMIB_to_OID ($) {
   my($arg) = @_;
-  my($quote, $buf, $var, $code, $val, $tmp, $tmpv, $strt);
-  my($ret, $pass, $pos, $need2pass, $cnt, %prev);
-  my(%Link) = (
-    'org' => 'iso',
-    'dod' => 'org',
-    'internet' => 'dod',
-    'directory' => 'internet',
-    'mgmt' => 'internet',
-    'mib-2' => 'mgmt',
-    'experimental' => 'internet',
-    'private' => 'internet',
-    'enterprises' => 'private',
-  );
+  my($cnt, $quote, $buf, %tOIDs, $tgot);
+  my($var, @parts, $strt, $indx, $ind, $val);
 
   if (!open(MIB, $arg)) {
     carp "snmpMIB_to_OID: Can't open $arg: $!"
@@ -1164,151 +1164,148 @@ sub snmpMIB_to_OID ($) {
     return -1;
   }
   print "snmpMIB_to_OID: loading $arg\n" if $SNMP_util::Debug;
-  $ret = 0;
-  $pass = 0;
-  $need2pass = 1;
   $cnt = 0;
-  $pos = tell(MIB);
-  while($need2pass) {
-    while(<MIB>) {
-      s/--.*--//g;	# throw away comments (-- anything --)
-      s/--.*//;		# throw away comments (-- anything EOL)
-      if ($quote) {
-	next unless /"/;
-	$quote = 0;
-      }
-      chomp;
-#
-#	$buf = "$buf $_";
-# Previous line removed (and following replacement)
-# suggested by Brian Reichert, reichert@numachi.com
-#
-      $buf .= ' ' . $_;
-      $buf =~ s/\s+/ /g;
-
-      if ($buf =~ / DEFINITIONS ::= BEGIN/) {
-	if ($pass == 0 && $need2pass) {
-	  seek(MIB, $pos, 0);
-	  $buf = "";
-	  $pass = 1;
-	  $need2pass = 0;
-	  $cnt = 0;
-	  next;
-	}
-	$need2pass = 0;
-	$pass = 0;
-	$pos = tell(MIB);
-	undef %Link;
-	undef %prev;
-	%Link = (
-	  'org' => 'iso',
-	  'dod' => 'org',
-	  'internet' => 'dod',
-	  'directory' => 'internet',
-	  'mgmt' => 'internet',
-	  'mib-2' => 'mgmt',
-	  'experimental' => 'internet',
-	  'private' => 'internet',
-	  'enterprises' => 'private',
-	);
-	$buf = "";
-	next;
-      }
-
-      $buf =~ s/OBJECT-TYPE/OBJECT IDENTIFIER/;
-      $buf =~ s/OBJECT-IDENTITY/OBJECT IDENTIFIER/;
-      $buf =~ s/OBJECT-GROUP/OBJECT IDENTIFIER/;
-      $buf =~ s/MODULE-IDENTITY/OBJECT IDENTIFIER/;
-      $buf =~ s/ IMPORTS .*\;//;
-      $buf =~ s/ SEQUENCE {.*}//;
-      $buf =~ s/ SYNTAX .*//;
-      $buf =~ s/ [\w\-]+ ::= OBJECT IDENTIFIER//;
-      $buf =~ s/ OBJECT IDENTIFIER .* ::= {/ OBJECT IDENTIFIER ::= {/;
-      $buf =~ s/".*"//;
-      if ($buf =~ /"/) {
-	$quote = 1;
-      }
-
-      if ($buf =~ / ([\w\-]+) OBJECT IDENTIFIER ::= {([^}]+)}/) {
-	$var = $1;
-	$buf = $2;
-	undef $val;
-	$buf =~ s/ +$//;
-	($code, $val) = split(' ', $buf, 2);
-
-	if (!defined($val) || (length($val) <= 0)) {
-	  $SNMP_util::OIDS{$var} = $code;
-	  $cnt++;
-	  print "'$var' => '$code'\n" if $SNMP_util::Debug;
-	} else {
-	  $strt = $code;
-	  while($val =~ / /) {
-	    ($tmp, $val) = split(' ', $val, 2);
-	    if ($tmp =~ /([\w\-]+)\((\d+)\)/) {
-	      $tmp = $1;
-	      if (exists($SNMP_util::OIDS{$strt})) {
-		$tmpv = "$SNMP_util::OIDS{$strt}.$2";
-	      } else {
-		$tmpv = $2;
-	      }
-	      $Link{$tmp} = $strt;
-	      if (!exists($prev{$tmp}) && exists($SNMP_util::OIDS{$tmp})) {
-		if ($tmpv ne $SNMP_util::OIDS{$tmp}) {
-		  $strt = "$strt.$tmp";
-		  $SNMP_util::OIDS{$strt} = $tmpv;
-		  $cnt++;
-		}
-	      } else {
-		$prev{$tmp} = 1;
-		$SNMP_util::OIDS{$tmp} = $tmpv;
-		$cnt++;
-		$strt = $tmp;
-	      }
-	    }
-	  }
-
-	  if (!exists($SNMP_util::OIDS{$strt})) {
-	    if ($pass) {
-	      carp "snmpMIB_to_OID: $arg: \"$strt\" prefix unknown, load the parent MIB first.\n"
-		unless ($SNMP_Session::suppress_warnings > 1);
-	    } else {
-		$need2pass = 1;
-	    }
-	  }
-	  $Link{$var} = $strt;
-	  if (exists($SNMP_util::OIDS{$strt})) {
-	    $val = "$SNMP_util::OIDS{$strt}.$val";
-	  }
-	  if (!exists($prev{$var}) && exists($SNMP_util::OIDS{$var})) {
-	    if ($val ne $SNMP_util::OIDS{$var}) {
-	      $var = "$strt.$var";
-	    }
-	  }
-
-	  $SNMP_util::OIDS{$var} = $val;
-	  $prev{$var} = 1;
-	  $cnt++;
-
-	  print "'$var' => '$val'\n" if $SNMP_util::Debug;
-	}
-	undef $buf;
-      }
-    }
-    if ($pass == 0 && $need2pass) {
-      seek(MIB, $pos, 0);
-      $buf = "";
-      $pass = 1;
-      $cnt = 0;
+  $quote = 0;
+  $tgot = 0;
+  $buf = '';
+  while(<MIB>) {
+    if ($quote) {
+      next unless /"/;
+      $quote = 0;
     } else {
-      $ret += $cnt;
-      $need2pass = 0;
+	s/--.*--//g;		# throw away comments (-- anything --)
+	s/^\s*--.*//;		# throw away comments at start of line
+    }
+    chomp;
+
+    $buf .= ' ' . $_;
+
+    $buf =~ s/"[^"]*"//g;
+    if ($buf =~ /"/) {
+      $quote = 1;
+      next;
+    }
+    $buf =~ s/--.*--//g;	# throw away comments (-- anything --)
+    $buf =~ s/--.*//;		# throw away comments (-- anything EOL)
+    $buf =~ s/\s+/ /g;
+    if ($buf =~ /DEFINITIONS *::= *BEGIN/) {
+	$cnt += MIB_fill_OID(\%tOIDs) if ($tgot);
+	$buf = '';
+	%tOIDs = ();
+	$tgot = 0;
+	next;
+    }
+    $buf =~ s/OBJECT-TYPE/OBJECT IDENTIFIER/;
+    $buf =~ s/OBJECT-IDENTITY/OBJECT IDENTIFIER/;
+    $buf =~ s/OBJECT-GROUP/OBJECT IDENTIFIER/;
+    $buf =~ s/MODULE-IDENTITY/OBJECT IDENTIFIER/;
+    $buf =~ s/ IMPORTS .*\;//;
+    $buf =~ s/ SEQUENCE *{.*}//;
+    $buf =~ s/ SYNTAX .*//;
+    $buf =~ s/ [\w\-]+ *::= *OBJECT IDENTIFIER//;
+    $buf =~ s/ OBJECT IDENTIFIER.*::= *{/ OBJECT IDENTIFIER ::= {/;
+
+    if ($buf =~ / ([\w\-]+) OBJECT IDENTIFIER *::= *{([^}]+)}/) {
+      $var = $1;
+      $buf = $2;
+      $buf =~ s/ +$//;
+      $buf =~ s/\s+\(/\(/g;	# remove spacing around '('
+      $buf =~ s/\(\s+/\(/g;
+      $buf =~ s/\s+\)/\)/g;	# remove spacing before ')'
+      @parts = split(' ', $buf);
+      $strt = '';
+      foreach $indx (@parts) {
+	if ($indx =~ /([\w\-]+)\((\d+)\)/) {
+	  $ind = $1;
+	  $val = $2;
+	  if (exists($tOIDs{$strt})) {
+	    $tOIDs{$ind} = $tOIDs{$strt} . '.' . $val;
+	  } elsif ($strt ne '') {
+	    $tOIDs{$ind} = "${strt}.${val}";
+	  } else {
+	    $tOIDs{$ind} = $val;
+	  }
+	  $strt = $ind;
+	  $tgot = 1;
+	} elsif ($indx =~ /^\d+$/) {
+	  if (exists($tOIDs{$strt})) {
+	    $tOIDs{$var} = $tOIDs{$strt} . '.' . $indx;
+	  } else {
+	    $tOIDs{$var} = "${strt}.${indx}";
+	  }
+	  $tgot = 1;
+	} else {
+	  $strt = $indx;
+	}
+      }
+      $buf = '';
     }
   }
-  close(MIB);
-  $RevNeeded = 1;
-  return $ret;
+  $cnt += MIB_fill_OID(\%tOIDs) if ($tgot);
+  $RevNeeded = 1 if ($cnt > 0);
+  return $cnt;
 }
 
+#
+# Fill the OIDS hash with results from the MIB parsing
+#
+sub MIB_fill_OID ($)
+{
+  my($href) = @_;
+  my($cnt, $changed, @del, $var, $val, @parts, $indx);
+  my(%seen);
+
+  $cnt = 0;
+  do {
+    $changed = 0;
+    @del = ();
+    foreach $var (keys %$href) {
+      $val = $href->{$var};
+      @parts = split('\.', $val);
+      $val = '';
+      foreach $indx (@parts) {
+	if ($indx =~ /^\d+$/) {
+	  $val .= '.' . $indx;
+	} else {
+	  if (exists($SNMP_util::OIDS{$indx})) {
+	    $val = $SNMP_util::OIDS{$indx};
+	  } else {
+	    $val .= '.' . $indx;
+	  }
+	}
+      }
+      if ($val =~ /^[\d\.]+$/) {
+	$val =~ s/^\.//;
+	if (!exists($SNMP_util::OIDS{$var})
+	|| (length($val) > length($SNMP_util::OIDS{$var}))) {
+	  $SNMP_util::OIDS{$var} = $val;
+	  print "'$var' => '$val'\n" if $SNMP_util::Debug;
+	  $changed = 1;
+	  $cnt++;
+	}
+	push @del, $var;
+      }
+    }
+    foreach $var (@del) {
+      delete $href->{$var};
+    }
+  } while($changed);
+
+  $Carp::CarpLevel++;
+  foreach $var (sort keys %$href) {
+    $val = $href->{$var};
+    $val =~ s/\..*//;
+    next if (exists($seen{$val}));
+    $seen{$val} = 1;
+    $seen{$var} = 1;
+    carp "snmpMIB_to_OID: prefix \"$val\" unknown, load the parent MIB first.\n"
+      unless ($SNMP_Session::suppress_warnings > 1);
+  }
+  $Carp::CarpLevel--;
+  return $cnt;
+}
+
+
 sub encode_oid_with_errmsg ($) {
   my ($oid) = @_;
   my $tmp = encode_oid(split(/\./, $oid));
